#include <avr/io.h>
//#include <avr/interrupt.h>
#include <util/delay.h>

#include "drv_twi.h"


#ifdef DEBUG
#include <stdio.h>
#define TWI_Debug(x)					printf x
#define TWI_DebugError(x)				printf x
#else
#define TWI_Debug(x)
#define TWI_DebugError(x)
#endif


static uint32_t twi_ready(TWI_t *);
static uint32_t twi_repeatstart(TWI_t *);
static uint32_t twi_send_dev_addr(TWI_t *, uint8_t);
static uint32_t twi_send_data(TWI_t *, uint8_t);
static uint32_t twi_read_ack_data(TWI_t *, uint8_t *);
static uint32_t twi_read_nack_data(TWI_t *, uint8_t *);
static uint32_t twi_stop(TWI_t *);




uint32_t TWI_Initialize(TWI_t *ptwi, uint32_t fperi_clk, uint32_t twi_speed)
{
	//SDASETUP 4CYC; SDAHOLD OFF; FMPEN disabled;
	ptwi->CTRLA = 0x00;

	//Debug Run
	ptwi->DBGCTRL = 0x00;

	//Master Baud Rate Control
	// Formula from the datasheet :
	// fSCL = fCLK_PER / (10 + 2MBAUD + fCLK_PER * Trise)
	// => 10 + 2MBAUD + fCLK_PER * Trise = fCLK_PER / fSCL
	// => 2MBAUD = (fCLK_PER / fSCL) - 10 - fCLK_PER * Trise
	// => MBAUD = ((fCLK_PER / fSCL) - 10 - fCLK_PER * Trise) / 2

	ptwi->MBAUD = (uint8_t)((((float)fperi_clk / twi_speed) - 10 - (float)fperi_clk * 0 / 1000000) / 2. + 0.5);

	//RIEN disabled; WIEN disabled; QCEN disabled; TIMEOUT DISABLED; SMEN disabled; ENABLE enabled;
	ptwi->MCTRLA = 0x01;

	//RIF disabled; WIF disabled; CLKHOLD disabled; ARBLOST disabled; BUSERR disabled; BUSSTATE UNKNOWN;

	//FLUSH disabled; ACKACT ACK; MCMD NOACT;


	return 0;
}




#ifdef __APPLY_TWI_TIMEOUT__
static uint32_t twi_ready(TWI_t *ptwi)
{
	uint32_t time_out = 1000;


	/* Do flush and set IDLE state */
	ptwi->MCTRLB = (1 << TWI_FLUSH_bp);
	ptwi->MSTATUS |= TWI_BUSSTATE_IDLE_gc;

	while(time_out > 0)
	{
		/* Check current job finish */
		if((ptwi->MSTATUS & TWI_BUSSTATE_BUSY_gc) != TWI_BUSSTATE_BUSY_gc)
		{
			break;
		}

		time_out--;
		_delay_us(10);
	}

	if(time_out == 0)
		return 1;


	return 0;
}
#else
static uint32_t twi_ready(TWI_t *ptwi)
{
	/* Do flush and set IDLE state */
	ptwi->MCTRLB = (1 << TWI_FLUSH_bp);
	ptwi->MSTATUS |= TWI_BUSSTATE_IDLE_gc;

	/* Check busy condition */
	while((ptwi->MSTATUS & TWI_BUSSTATE_BUSY_gc) == TWI_BUSSTATE_BUSY_gc)
	{
		_delay_us(10);
	}


	return 0;
}
#endif




#ifdef __APPLY_TWI_TIMEOUT__
static uint32_t twi_repeatstart(TWI_t *ptwi)
{
	uint32_t time_out = 1000;


	ptwi->MCTRLB &= 0xFC;
	ptwi->MCTRLB |= TWI_MCMD_REPSTART_gc;

	while(time_out > 0)
	{
		/* Check busy condition */
		if((ptwi->MSTATUS & TWI_BUSSTATE_BUSY_gc) != TWI_BUSSTATE_BUSY_gc)
		{
			break;
		}

		time_out--;
		_delay_us(10);
	}

	if(time_out == 0)
		return 1;


	return 0;
}
#else
static uint32_t twi_repeatstart(TWI_t *ptwi)
{
	ptwi->MCTRLB &= 0xFC;
	ptwi->MCTRLB |= TWI_MCMD_REPSTART_gc;

	/* Check busy condition */
	while((ptwi->MSTATUS & TWI_BUSSTATE_BUSY_gc) == TWI_BUSSTATE_BUSY_gc)
	{
		_delay_us(10);
	}

	return 0;
}
#endif




#ifdef __APPLY_TWI_TIMEOUT__
static uint32_t twi_send_dev_addr(TWI_t *ptwi, uint8_t addr)
{
	uint32_t time_out = 1000;


	/* Send Master Address */
	ptwi->MADDR = addr;

	/* For master write, the code will know that no acknowledge should be sent
	since it is itself sending data. */
	ptwi->MCTRLB |= TWI_ACKACT_NACK_gc;

	while(time_out > 0)
	{
		/* Check whether status is unknown */
		if((ptwi->MSTATUS & TWI_BUSSTATE_gm) == 0x0)
		{
			return 1;
		}

		/* Device address has attribute of "READ", so RIF also has to be read */
		if((ptwi->MSTATUS & TWI_RIF_bm) || (ptwi->MSTATUS & TWI_WIF_bm))
		{
			/* Check ACK condition */
			if((ptwi->MSTATUS & TWI_RXACK_bp) != TWI_RXACK_bp)
			{
				break;
			}
		}

		time_out--;
		_delay_us(10);
	}

	if(time_out == 0)
		return 1;


	return 0;
}
#else
static uint32_t twi_send_dev_addr(TWI_t *ptwi, uint8_t addr)
{
	/* Send Master Address */
	ptwi->MADDR = addr;

	/* For master write, the code will know that no acknowledge should be sent
	since it is itself sending data. */
	ptwi->MCTRLB |= TWI_ACKACT_NACK_gc;

		/* Check whether status is unknown */
	if((ptwi->MSTATUS & TWI_BUSSTATE_gm) == 0x0)
	{
		return 1;
	}

	/* Check current job finish */
	while(!(ptwi->MSTATUS & TWI_RIF_bm) && !(ptwi->MSTATUS & TWI_WIF_bm))
	{
		_delay_us(10);
	}

	/* Check ACK condition */
	while((ptwi->MSTATUS & TWI_RXACK_bp) == TWI_RXACK_bp)
	{
		_delay_us(10);
	}


	return 0;
}
#endif




#ifdef __APPLY_TWI_TIMEOUT__
static uint32_t twi_send_data(TWI_t *ptwi, uint8_t data)
{
	uint32_t time_out = 1000;


	/* Send data */
	ptwi->MDATA = data;

	/* For master write, the code will know that no acknowledge should be sent
	since it is itself sending data. */
	ptwi->MCTRLB |= TWI_ACKACT_NACK_gc;

	while(time_out > 0)
	{
		/* Check current job finish */
	   //if ((ptwi->MSTATUS & TWI_RIF_bm) || (ptwi->MSTATUS & TWI_WIF_bm))
	   if(ptwi->MSTATUS & TWI_WIF_bm)
	   {
			/* Check ACK condition */
			if((ptwi->MSTATUS & TWI_RXACK_bp) != TWI_RXACK_bp)
			{
				break;
			}
	   }

		time_out--;_delay_us(10);
	}

	if(time_out == 0)
		return 1;


	return 0;
}
#else
static uint32_t twi_send_data(TWI_t *ptwi, uint8_t data)
{
	/* Send data */
	ptwi->MDATA = data;

	/* For master write, the code will know that no acknowledge should be sent
	since it is itself sending data. */
	ptwi->MCTRLB |= TWI_ACKACT_NACK_gc;

	/* Check current job finish */
	//while(!(ptwi->MSTATUS & TWI_RIF_bm) && !(ptwi->MSTATUS & TWI_WIF_bm))
	while(!(ptwi->MSTATUS & TWI_WIF_bm))
	{
		_delay_us(10);
	}

	/* Check ACK condition */
	while((ptwi->MSTATUS & TWI_RXACK_bp) == TWI_RXACK_bp)
	{
		_delay_us(10);
	}


	return 0;
}
#endif




#ifdef __APPLY_TWI_TIMEOUT__
static uint32_t twi_read_ack_data(TWI_t *ptwi, uint8_t *pdata)
{
	uint32_t time_out = 1000;


	while(time_out > 0)
	{
		/* Check current job finish */
		if((ptwi->MSTATUS & TWI_RIF_bm) == TWI_RIF_bm)
		{
			/* Read data */
			*pdata = ptwi->MDATA;

			ptwi->MSTATUS |= TWI_RIF_bm;
			ptwi->MCTRLB |= TWI_MCMD_RECVTRANS_gc;
			break;
		}

		time_out--;
		_delay_us(10);
	}

	ptwi->MCTRLB &= ~(1 << TWI_ACKACT_bp);

	if(time_out == 0)
		return 1;

	return 0;
}
#else
static uint32_t twi_read_ack_data(TWI_t *ptwi, uint8_t *pdata)
{
	while(1)
	{
		/* Check current job finish */
		if((ptwi->MSTATUS & TWI_RIF_bm) == TWI_RIF_bm)
		{
			/* Read data */
			*pdata = ptwi->MDATA;

			ptwi->MSTATUS |= TWI_RIF_bm;
			ptwi->MCTRLB |= TWI_MCMD_RECVTRANS_gc;
			break;
		}

		_delay_us(10);
	}

	ptwi->MCTRLB &= ~(1 << TWI_ACKACT_bp);

	return 0;
}
#endif




#ifdef __APPLY_TWI_TIMEOUT__
static uint32_t twi_read_nack_data(TWI_t *ptwi, uint8_t *pdata)
{
	uint32_t time_out = 1000;

	while(time_out > 0)
	{
		/* Check current job finish */
		if((ptwi->MSTATUS & TWI_RIF_bm) == TWI_RIF_bm)
		{
			/* Read data */
			*pdata = ptwi->MDATA;

			ptwi->MSTATUS |= TWI_RIF_bm;
			break;
		}

		time_out--;
		_delay_us(10);
	}

	ptwi->MCTRLB |= TWI_ACKACT_NACK_gc;

	if(time_out == 0)
		return 1;


	return 0;
}
#else
static uint32_t twi_read_nack_data(TWI_t *ptwi, uint8_t *pdata)
{
	while(1)
	{
		/* Check current job finish */
		if((ptwi->MSTATUS & TWI_RIF_bm) == TWI_RIF_bm)
		{
			/* Read data */
			*pdata = ptwi->MDATA;

			ptwi->MSTATUS |= TWI_RIF_bm;
			break;
		}

		_delay_us(10);
	}

	ptwi->MCTRLB |= TWI_ACKACT_NACK_gc;


	return 0;
}
#endif




#ifdef __APPLY_TWI_TIMEOUT__
static uint32_t twi_stop(TWI_t *ptwi)
{
	uint32_t time_out = 1000;


	ptwi->MCTRLB &= 0xFC;
	ptwi->MCTRLB |= TWI_MCMD_STOP_gc;

	while(time_out > 0)
	{
		/* Check busy condition */
		if((ptwi->MSTATUS & TWI_BUSSTATE_BUSY_gc) != TWI_BUSSTATE_BUSY_gc)
		{
			break;
		}

		time_out--;
		_delay_us(10);
	}

	if(time_out == 0)
		return 1;


	return 0;

}
#else
static uint32_t twi_stop(TWI_t *ptwi)
{
	ptwi->MCTRLB &= 0xFC;
	ptwi->MCTRLB |= TWI_MCMD_STOP_gc;

	/* Check busy condition */
	while((ptwi->MSTATUS & TWI_BUSSTATE_BUSY_gc) == TWI_BUSSTATE_BUSY_gc)
	{
		_delay_us(10);
	}

	return 0;
}
#endif




uint32_t TWI_Write(TWI_t *ptwi, uint8_t dev_addr, uint8_t addr_size, uint16_t reg_addr, uint8_t *pbuf, uint32_t size)
{
	uint8_t write_flag = 0;


	if(twi_ready(ptwi))
	{
		//MAIN_DebugError(("# start error\r\n"));
		return 1;
	}

	if(twi_send_dev_addr(ptwi, (dev_addr << 1) | write_flag))
	{
		//MAIN_DebugError(("# twi_send_dev_addr error\r\n"));
		return 2;
	}

//#ifdef __DEVICE_HAS_REG_ADDR_16BITS__
	if(addr_size == REG_ADDR_BITSIZE_16)
	{
		if(twi_send_data(ptwi, (uint8_t)(reg_addr >> 8)))
		{
			//MAIN_DebugError(("# twi_send_data #1 error\r\n"));
			return 3;
		}
	}
//#endif

	if(twi_send_data(ptwi, (uint8_t)(reg_addr & 0xFF)))
	{
		//MAIN_DebugError(("# twi_send_data #2 error\r\n"));
		return 4;
	}

	for(uint32_t count = 0; count < size; count++)
	{
		if(twi_send_data(ptwi, *pbuf))
		{
			//MAIN_DebugError(("Error twi_send_data at %d\r\n", count));
			twi_stop(ptwi);

			return 5;
		}

		pbuf++;
	}

	twi_stop(ptwi);


	return 0;
}




uint32_t TWI_Read(TWI_t *ptwi, uint8_t dev_addr, uint8_t addr_size, uint16_t reg_addr, uint8_t *pdata, uint32_t size)
{
	uint8_t read_flag = 0;


	if(twi_ready(ptwi))
		return 1;

	if(twi_send_dev_addr(ptwi, (dev_addr << 1) | read_flag))
		return 2;

//#ifdef __DEVICE_HAS_REG_ADDR_16BITS__
	if(addr_size == REG_ADDR_BITSIZE_16)
	{
		if(twi_send_data(ptwi, (uint8_t)(reg_addr >> 8)))
			return 3;
	}
//#endif

	if(twi_send_data(ptwi, (uint8_t)(reg_addr & 0xFF)))
		return 4;

    if(twi_repeatstart(ptwi))
	{
		//MAIN_DebugError(("twi_repeatstart error\r\n"));
	    return 5;
	}

	read_flag = 1;

    if(twi_send_dev_addr(ptwi, (dev_addr << 1) | read_flag))
		return 6;

    for(uint32_t count = 0;  count < size - 1; count++)
    {
	    if(twi_read_ack_data(ptwi, pdata))
		{
			//MAIN_DebugError("twi_read_ack_data() error\r\n"));
			twi_stop(ptwi);
		    return 7;
		}

		pdata++;
    }

    if(twi_read_nack_data(ptwi, pdata))
	{
		//MAIN_DebugError(("twi_read_nack_data() error\r\n"));
		twi_stop(ptwi);
	    return 8;
	}

    twi_stop(ptwi);


	return 0;
}




#ifdef __APPLY_I2C_SW_RESET__
void Do_SW_Reset(TWI_t *ptwi)
{
	twi_ready(ptwi);

	twi_send_dev_addr(ptwi, 0xFF);
	
	twi_repeatstart(ptwi);

	twi_stop(ptwi);
}
#endif

/* End of File */
