#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "atmega4809_xplained_pro.h"

#include "drv_twi.h"
#include "drv_usart.h"


#ifdef DEBUG
#include <stdio.h>
#define MAIN_Debug(x)					printf x
#define MAIN_DebugError(x)				printf x
#else
#define MAIN_Debug(x)
#define MAIN_DebugError(x)
#endif


// [HT16K33 LED Controller device address]
#define LED_DEVICE_ADDR					(0xE0 >> 1)

/* Gap Time */
#define GAP_TIME						10


static uint32_t Seg_SystemSetup(uint8_t osc_flag);
static uint32_t Seg_CtrlDisplay(uint8_t flag);
static uint32_t Seg_CtrlDimming(uint8_t level);
static uint32_t Seg_CtrlColon(uint8_t flag);
static uint32_t Seg_ClearSegment(uint8_t digit_num);
static uint32_t Seg_SetSegment(uint8_t digit_num, uint8_t value, uint8_t dot_flag);

static void Display_Segment(uint16_t value);


static const uint8_t s_valuetable[] = {
	0x3F, /* 0 */
	0x06, /* 1 */
	0x5B, /* 2 */
	0x4F, /* 3 */
	0x66, /* 4 */
	0x6D, /* 5 */
	0x7D, /* 6 */
	0x07, /* 7 */
	0x7F, /* 8 */
	0x6F, /* 9 */
	0x77, /* a */
	0x7C, /* b */
	0x39, /* C */
	0x5E, /* d */
	0x79, /* E */
	0x71, /* F */
};




static void sysclk_init(void)
{
	//CPU_CCP = CCP_IOREG;
	CPU_CCP = 0xD8;

	CLKCTRL.MCLKCTRLA = CLKCTRL_CLKSEL_OSC20M_gc;

	//CPU_CCP = CCP_IOREG;
	CPU_CCP = 0xD8;

	//CLKCTRL.MCLKCTRLB = 0x1;
	CLKCTRL.MCLKCTRLB = 0x0;	//the Main Clock will pass through undivided (CLK_PER=CLK_MAIN), regardless of the value of PDIV
}




static void board_init()
{
	/* LED0 - Negative logic operation */
	PORTB.DIRSET = PIN_LED0;
	PORTB.OUTSET = PIN_LED0;

	/* User Button - Negative logic operation */
	PORTB.DIRCLR = PIN_SW0;

	PORTB.PIN2CTRL = PORT_PULLUPEN_bm;
	PORTB.PIN2CTRL &= ~0x7;
}




static void TWI0_Pin_ConfigurePin(void)
{
	PORTMUX.TWISPIROUTEA = 0x20;

	PORTC.PIN2CTRL = 0x00;
	PORTC.PIN3CTRL = 0x00;
}




void LED_Toggle(uint8_t pin)
{
	PORTB.OUTTGL = pin;
}




/**
 * \brief Application entry point for I2C 7-segment example.
 *
 * \return Unused (ANSI-C compatibility).
 */
int main(void)
{
	uint32_t count;


	sysclk_init();

	board_init();

	// Initialization of peripherals
	USART1_Printf_Init(115200);
	
	/* Needs delay to settle USART */
	//_delay_ms(500);

	MAIN_Debug(("****************************************************\r\n"));
	MAIN_Debug(("ATmega4809 Xplained Pro - TWI 2-digit 7-segment test\r\n"));
	MAIN_Debug(("****************************************************\r\n\n"));

	/* TWI0 configuration : use PC2 for SDA and PC3 for SCL */
	TWI0_Pin_ConfigurePin();

	TWI_Initialize(&TWI0, F_CPU, 400000);

	/* Wait at least 10 ms */
	_delay_ms(GAP_TIME);

	Seg_SystemSetup(1);

	Seg_CtrlDisplay(1);

	Seg_CtrlDimming(5);

	Seg_ClearSegment(0);
	Seg_ClearSegment(1);
	Seg_ClearSegment(2);
	Seg_ClearSegment(3);
	Seg_ClearSegment(4);

	count = 0;

	while (1)
	{
		//LED_Toggle(PIN_LED0);

		Display_Segment(count);

		if(count == 10000)
		{
			count = 0;

			Seg_ClearSegment(0);
			Seg_ClearSegment(1);
			Seg_ClearSegment(3);
		}

		if(count > 1000)
			_delay_ms(1);
		else if(count > 100)
			_delay_ms(10);
		else if(count > 10)
			_delay_ms(100);
		else
			_delay_ms(200);

		count++;
	}
}




static uint32_t Seg_SystemSetup(uint8_t osc_flag)
{
	uint8_t tx_buf = 0x20 | osc_flag;
	uint32_t error_code;


	/* Write buffer to slave until success. */
	if((error_code = TWI_Write(&TWI0, LED_DEVICE_ADDR, REG_ADDR_BITSIZE_8, tx_buf, NULL, 0)) != 0)
	{
		printf("[L%d] %s() write ERROR - 0x%X\r\n", __LINE__, __FUNCTION__, error_code);
	}


	return error_code;
}




static uint32_t Seg_CtrlDisplay(uint8_t flag)
{
	uint8_t tx_buf = 0x80 | flag;
	uint32_t error_code;


	/* Write buffer to slave until success. */
	if((error_code = TWI_Write(&TWI0, LED_DEVICE_ADDR, REG_ADDR_BITSIZE_8, tx_buf, NULL, 0)) != 0)
	{
		printf("[L%d] %s() write ERROR - 0x%X\r\n", __LINE__, __FUNCTION__, error_code);
	}


	return error_code;
}




static uint32_t Seg_CtrlDimming(uint8_t level)
{
	uint8_t tx_buf = 0xE0 | level;
	uint32_t error_code;


	/* Write buffer to slave until success. */
	if((error_code = TWI_Write(&TWI0, LED_DEVICE_ADDR, REG_ADDR_BITSIZE_8, tx_buf, NULL, 0)) != 0)
	{
		printf("[L%d] %s() write ERROR - 0x%X\r\n", __LINE__, __FUNCTION__, error_code);
	}


	return error_code;
}




static uint32_t Seg_CtrlColon(uint8_t flag)
{
	uint8_t tx_buf[2];
	uint32_t error_code;


	tx_buf[0] = 0x04;
	tx_buf[1] = flag << 1;

	if((error_code = TWI_Write(&TWI0, LED_DEVICE_ADDR, REG_ADDR_BITSIZE_8, tx_buf[0], &tx_buf[1], 1)) != 0)
	{
		printf("[L%d] %s() write ERROR - 0x%X\r\n", __LINE__, __FUNCTION__, error_code);
	}


	return error_code;
}




static uint32_t Seg_ClearSegment(uint8_t digit_num)
{
	uint8_t tx_buf[2];
	uint32_t error_code;


	tx_buf[0] = digit_num * 2;
	tx_buf[1] = 0;

	/* Write buffer to slave until success. */
	if((error_code = TWI_Write(&TWI0, LED_DEVICE_ADDR, REG_ADDR_BITSIZE_8, tx_buf[0], &tx_buf[1], 1)) != 0)
	{
		printf("[L%d] %s() write ERROR - 0x%X\r\n", __LINE__, __FUNCTION__, error_code);
	}


	return error_code;
}




static uint32_t Seg_SetSegment(uint8_t digit_num, uint8_t value, uint8_t dot_flag)
{
	uint8_t tx_buf[2];
	uint32_t error_code;


	tx_buf[0] = digit_num * 2;
	tx_buf[1] = s_valuetable[value] | (dot_flag << 7);

	/* Write buffer to slave until success. */
	if((error_code = TWI_Write(&TWI0, LED_DEVICE_ADDR, REG_ADDR_BITSIZE_8, tx_buf[0], &tx_buf[1], 1)) != 0)
	{
		printf("[L%d] %s() write ERROR - 0x%X\r\n", __LINE__, __FUNCTION__, error_code);
	}


	return error_code;
}




static void Display_Segment(uint16_t value)
{
	uint8_t dig0_value = (value / 1000);
	uint8_t dig1_value = (value % 1000) / 100;
	uint8_t dig3_value = (value % 100) / 10;
	uint8_t dig4_value = (value % 10);

	if(value >= 1000)
		Seg_SetSegment(0, dig0_value, 0);

	if(value >= 100)
		Seg_SetSegment(1, dig1_value, 0);

	if(value >= 10)
		Seg_SetSegment(3, dig3_value, 0);

	Seg_SetSegment(4, dig4_value, 0);
}

/* End of File */
